@@ -28,7 +28,7 @@ from virt_up.instance import Instance, Settings
 usage="""\
 virt-up [--name] <name> [--template <template>] [options]
                --list [--all] | --list-templates
-               --login [--name] <name> [--command "<command>"]
+               --login [--name] <name> [--sftp] [--command "<command>"]
                --delete [--name] <name> | --delete --all
 """
 
@@ -55,13 +55,13 @@ def list_instances(args):
     """
     List instances.
     """
-    clones_only = (not args.all)
-    for name in sorted(Instance.list(clones_only)):
-        sys.stdout.write(f'{name}\n')
+    for instance in Instance.all():
+        if args.all or not instance.is_template():
+            sys.stdout.write(f'{instance.name}\n')
 
 def list_templates(args):
     """
-    List available templates.
+    List available template definitions.
     """
     settings = Settings()
     for name in sorted(settings.templates.keys()):
@@ -71,28 +71,39 @@ def list_templates(args):
 
 def delete(args):
     """
-    Delete the instance.
+    Delete the instance or all of the instances.
     """
     if not (bool(args.name) ^ bool(args.all)):
         die(f'<name> or --all is required.\nusage: {usage}')
-    if args.name:
-        if Instance.exists(args.name):
-            instance = Instance(args.name)
-            instance.delete()
-    else:
-        names = list(Instance.list(clones_only=False))
-        if names:
-            if args.yes:
-                answer = 'yes'
-            else:
-                names_str = ', '.join(names)
-                sys.stdout.write(f'About to delete instances: {names_str}\n')
-                answer = input('Continue? [y/n] > ').lower()
-            if answer in ('y', 'yes'):
-                for name in names:
-                    if Instance.exists(name):
-                        instance = Instance(name)
-                        instance.delete()
+
+    if not args.name:
+        delete_all(args)
+    elif Instance.exists(args.name):
+        instance = Instance(args.name)
+        instance.delete()
+
+def delete_all(args):
+    # Remove clones first, then templates.
+    clones = []
+    templates = []
+    for instance in Instance.all():
+        if instance.is_template():
+            templates.append(instance)
+        else:
+            clones.append(instance)
+    if not args.yes:
+        names = []
+        names.extend([i.name for i in clones])
+        names.extend([i.name for i in templates])
+        names = ', '.join(names)
+        sys.stdout.write(f'About to delete: {names}\n')
+        answer = input('Continue? [y/n] > ').lower()
+        if answer not in ('y', 'yes'):
+            return # bail
+    for instance in clones:
+        instance.delete()
+    for instance in templates:
+        instance.delete()
 
 def login(args):
     """
@@ -102,9 +113,14 @@ def login(args):
         die(f'<name> is required.\nusage: {usage}')
     if not Instance.exists(args.name):
         log.error(f"Instance '{args.name}' not found.")
-    else:
-        instance = Instance(args.name)
-        instance.login(args.command)
+        return
+    options = {}
+    if args.sftp:
+        options['mode'] = 'sftp'
+    if args.command:
+        options['command'] = args.command
+    instance = Instance(args.name)
+    instance.login(**options)
 
 def create(args):
     """
@@ -112,36 +128,27 @@ def create(args):
     the template to create a new instance. Just start the instance if it
     already exists.
     """
-    if not args.name:
-        die(f'<name> is required.\nusage: {usage}')
-    if args.template is None:
-        args.template = args.name
+    options = vars(args)
+    name = options.pop('name', None)
+    template = options.pop('template', name)
+    prefix = options.pop('prefix', 'TEMPLATE-')
 
-    options = {}
-    if args.root_password:
-        options['root_password'] = args.root_password
-    if args.user:
-        options['user'] = args.user
-    if args.password:
-        options['password'] = args.password
-    if args.size:
-        options['size'] = args.size
-    if args.memory:
-        options['memory'] = args.memory
-    if args.graphics:
-        options['graphics'] = args.graphics
+    if not (name or args.no_clone):
+        die(f'<name> is required.\nusage: {usage}')
 
     try:
-        template = Instance.build(args.template, prefix='TEMPLATE-', **options)
+        template = Instance.build(name=template, template=template, prefix=prefix, **options)
         if args.no_clone:
             return
-        instance = template.clone(args.name, **options)
+        instance = template.clone(name, **options)
         instance.wait_for_port(22)
         log.info(f"Instance '{instance.name}' is up.")
+    except KeyError as e:
+        raise(e)
     except ValueError as e:
-        die(e)
+        die('ValueError:' + str(e))
     except LookupError as e:
-        die(e)
+        die('LookupError: ' + str(e))
 
 def main():
     parser = argparse.ArgumentParser(prog='virt-up', usage=usage)
@@ -151,7 +158,7 @@ def main():
     group.add_argument('--list', action='store_true', help='list instances')
     group.add_argument('--list-templates', action='store_true', help='list template names')
     group.add_argument('--delete', action='store_true', help='delete the instance')
-    group.add_argument('--login', action='store_true', help=argparse.SUPPRESS)
+    group.add_argument('--login', action='store_true', help='login to a running instance')
 
     group = parser.add_mutually_exclusive_group()
     group.add_argument('name', metavar='<name>', nargs='?', help='instance name')
@@ -162,12 +169,16 @@ def main():
     parser.add_argument('--user', metavar='<user>', help='username (default: virt)')
     parser.add_argument('--password', metavar='<password>', help='password (default: random)')
     parser.add_argument('--size', metavar='<size>', help='instance disk size (default: image size)')
-    parser.add_argument('--memory', metavar='<memory>', help='instance memory (default: 512)')
-    parser.add_argument('--vcpus', metavar='<vcpus>', help='instance vcpus (default: 1)')
-    parser.add_argument('--graphics', metavar='<graphics>', help='instance graphics type (default: none)')
+    parser.add_argument('--memory', metavar='<memory>', help='instance memory (default: 512)', default=512)
+    parser.add_argument('--vcpus', metavar='<vcpus>', help='instance vcpus (default: 1)', default=1)
+    parser.add_argument('--graphics', metavar='<graphics>', help='instance graphics type (default: none)', default='none')
+    parser.add_argument('--dns-domain', metavar='<dns-domain>', help='dns domain name')
 
+    parser.add_argument('--sftp', action='store_true', help='--login with sftp')
     parser.add_argument('--command', metavar='<command>', help='--login ssh command')
     parser.add_argument('--no-clone', action='store_true', help='build template instance only')
+    parser.add_argument('--no-inventory', dest='inventory', action='store_false',
+                                          help='exclude instance from the virt-up ansible inventory file')
     parser.add_argument('--all', action='store_true', help='include template instances')
     parser.add_argument('--yes', action='store_true', help='answer yes to interactive questions')
     parser.add_argument('--quiet', action='store_true', help='show less output')
