@@ -24,7 +24,11 @@ libvirt-based hypervisor.
 """
 
 import configparser
+import crypt
 import datetime
+import fcntl
+import io
+import glob
 import json
 import logging
 import os
@@ -38,7 +42,7 @@ import xml.etree.ElementTree
 import sh
 import libvirt
 
-from .config import SETTINGS, TEMPLATES
+from .config import SETTINGS, TEMPLATES, SCRIPTS
 
 log = logging.getLogger(__name__)
 
@@ -75,11 +79,16 @@ def _adjust_sh_log():
 # Commands
 cp = sh.Command('cp').bake(_out=logout, _err=logerr)
 ssh = sh.Command('ssh')
+sftp = sh.Command('sftp')
 ssh_keygen = sh.Command('ssh-keygen').bake(_out=logout, _err=logerr)
 qemu_img = sh.Command('qemu-img').bake(_out=logout, _err=logerr)
 virt_builder = sh.Command('virt-builder').bake(_out=logout, _err=logerr)
 virt_install = sh.Command('virt-install').bake(_out=logout, _err=logerr)
 virt_sysprep = sh.Command('virt-sysprep').bake(_out=logout, _err=logerr)
+try:
+    ansible = sh.Command('ansible-playbook').bake(_out=logout, _err=logerr)
+except sh.CommandNotFound:
+    ansible = None
 
 # Avoid writing "domain not found" errors to the console.
 def _libvirt_callback(userdata, err):
@@ -101,6 +110,8 @@ class Settings:
         self.image_format = self.site.get('image-format', 'qcow2')
         self.dns_domain = self.site.get('dns-domain', '')
         self.address_source = self.site.get('address-source', 'agent')
+        self.template_playbook = self.site.get('template-playbook', '')
+        self.instance_playbook = self.site.get('instance-playbook', '')
 
         # Load template definitions.
         self.templates = self._load('templates', TEMPLATES)
@@ -114,6 +125,8 @@ class Settings:
                 raise LookupError(f"Template '{template}' not found in settings.")
             self.template = self.templates[template]
             self.template_name = template
+            self.template_playbook = self.template.get('template-playbook')
+            self.instance_playbook = self.template.get('instance-playbook')
             self.os_version = self.template.get('os-version')
             self.os_variant = self.template.get('os-variant')
             self.address_source = self.template.get('address-source', self.address_source)
@@ -147,6 +160,35 @@ class Settings:
             args.extend(shlex.split(args_str))
         return args
 
+class LockFile:
+    """
+    Interprocess lock file.
+    """
+    def _write(self, text):
+        self.fp.seek(0)
+        self.fp.truncate()
+        self.fp.write(text)
+        self.fp.flush()
+        self.fp.seek(0)
+
+    def __enter__(self):
+        if os.path.exists('/var/run/user/%d' % os.getuid()):
+            path = '/var/run/user/%d/virt-up.lock' % os.getuid()
+        else:
+            path = '/tmp/virt-up.lock'
+        log.debug("Waiting for lock")
+        self.fp = open(path, 'a+')
+        fcntl.flock(self.fp.fileno(), fcntl.LOCK_EX)
+        self._write(str(os.getpid())) # For troubleshooting.
+        log.debug("Obtained lock")
+
+    def __exit__(self, *exc):
+        log.debug("Releasing lock")
+        self._write('')
+        fcntl.flock(self.fp.fileno(), fcntl.LOCK_UN)
+        self.fp.close()
+        log.debug("Released lock")
+
 class Connection:
     """
     A libvirt connection context manager.
@@ -201,7 +243,6 @@ class Creds:
             ssh_keygen('-t', 'rsa', '-N', '', '-f', ssh_identity)
         return ssh_identity
 
-
 class MacAddresses:
     filename = f'{xdg_data_home}/virt-up/macaddrs.json'
 
@@ -301,6 +342,12 @@ class Instance:
         with os.fdopen(os.open(self.metafile, flags, 0o600), 'w') as fp:
             json.dump(self.meta, fp, indent=4)
 
+    def is_clone(self):
+        return 'cloned' in self.meta
+
+    def is_template(self):
+        return 'cloned' not in self.meta
+
     def mac(self):
         if self._mac is None:
             root = xml.etree.ElementTree.fromstring(self.domain.XMLDesc())
@@ -325,12 +372,19 @@ class Instance:
         """
         if not self.domain.isActive():
             log.info(f"Starting instance '{self.name}'.")
-            self.domain.create()
             for retries in range(120, -1, -1):
-                time.sleep(2)
+                try:
+                    self.domain.create()
+                except libvirt.libvirtError as e:
+                    if e.get_error_code() == libvirt.VIR_ERR_OPERATION_INVALID:
+                        pass  # domain is running
+                    else:
+                        raise e
                 if self.domain.isActive():
-                    break
-                log.debug(f"Waiting for running state; {retries} left.")
+                    return
+                if retries > 0:
+                    log.debug(f"Waiting for running state; {retries} left.")
+                    time.sleep(2)
             if not self.domain.isActive():
                 raise TimeoutError(f"Failed to start instance '{self.name}'.")
 
@@ -340,13 +394,19 @@ class Instance:
         """
         if self.domain.isActive():
             log.info(f"Stopping instance '{self.name}'.")
-            self.domain.shutdown()
             for retries in range(120, -1, -1):
-                time.sleep(2)
+                try:
+                    self.domain.shutdown()
+                except libvirt.libvirtError as e:
+                    if e.get_error_code() == libvirt.VIR_ERR_OPERATION_INVALID:
+                        pass  # domain is not running
+                    else:
+                        raise e
                 if not self.domain.isActive():
-                    break
-                self.domain.shutdown()
-                log.debug(f"Waiting for shutdown state; {retries} left.")
+                    return
+                if retries > 0:
+                    log.debug(f"Waiting for shutdown state; {retries} left.")
+                    time.sleep(2)
             if self.domain.isActive():
                 raise TimeoutError(f"Failed to stop instance '{self.name}'.")
 
@@ -354,6 +414,17 @@ class Instance:
         """
         Delete the instance, disk images, and instance meta data.
         """
+        in_use = []
+        for instance in Instance.all():
+            from_ = instance.meta.get('from', '')
+            format_ = instance.meta.get('image_format', '')
+            if from_ == self.name and format_ == 'qcow2':
+                in_use.append(instance.name)
+        if in_use:
+            in_use = ', '.join(["'%s'" %x for x in in_use])
+            log.error(f"Unable to delete '{self.name}'; in use by {in_use}.")
+            return 1
+
         log.info(f"Destroying instance '{self.name}'.")
         rm_f(self.metafile)
         self.meta = None
@@ -373,6 +444,7 @@ class Instance:
         self._disks = None
         self._mac = None
         self._address = None
+        Instance.update_inventory()
 
     def _ia_to_addresses(self, ia):
         """
@@ -482,6 +554,30 @@ class Instance:
                 time.sleep(2)
         return address
 
+    def _address_from_dns(self):
+        """
+        Attempt to retreive the instance address from dns. Assumes the dns is automatically updated by
+        the dhcp server with the hostname configured on the guests.
+        """
+        address = None
+        hostname = self.meta.get('hostname')
+        if not hostname:
+            raise LookupError(f"hostname is missing in instance '{self.name}' meta file.")
+        for retries in range(120, -1, -1):
+            try:
+                ai = socket.getaddrinfo(hostname, 22, family=socket.AF_INET, proto=socket.IPPROTO_TCP)
+                if ai:
+                    address = ai[0][4][0]
+                if address:
+                    break
+            except:
+                pass
+            if retries > 0:
+                suffix = 'ies' if retries > 1 else 'y'
+                log.debug(f"Waiting for instance '{self.name}' address from dns lookup; {retries} retr{suffix} left.")
+                time.sleep(2)
+        return address
+
     def address(self):
         """
         Get the public IPv4 address for login.
@@ -493,13 +589,15 @@ class Instance:
         if not self.domain.isActive():
             self.start()
 
-        address_source = self.meta.get('address_source', 'agent')
+        address_source = self.meta.get('address-source', 'agent')
         if address_source == 'agent':
             address = self._address_from_ia(source='agent')
         elif address_source == 'lease':
             address = self._address_from_ia(source='lease')
         elif address_source == 'arp':
             address = self._address_from_arp()
+        elif address_source == 'dns':
+            address = self._address_from_dns()
         else:
             raise ValueError(f"Invalid address_source '{address_source}' in instance '{self.name}'.")
 
@@ -518,9 +616,8 @@ class Instance:
                 s.settimeout(2)
                 s.connect((address, int(port)))
                 return True
-            except Exception as e:
-                if str(e) == "[Errno 111] Connection refused":
-                    time.sleep(2) 
+            except:
+                pass
             finally:
                 try:
                     s.shutdown(socket.SHUT_RDWR)
@@ -530,25 +627,16 @@ class Instance:
             if retries > 0:
                 suffix = 'ies' if retries > 1 else 'y'
                 log.debug(f"Waiting for open port '{port}' on address '{address}'; {retries} retr{suffix} left.")
+                time.sleep(2)
 
         raise LookupError(f"Unable to connect to '{address}:{port}'.")
 
     @classmethod
-    def list(cls, clones_only=True):
-        with Connection() as conn:
-            for domain in conn.listAllDomains():
-                name = domain.name()
-                metafile = f'{xdg_data_home}/virt-up/instance/{name}.json'
-                try:
-                    with open(metafile, 'r') as fp:
-                        meta = json.load(fp)
-                    if clones_only:
-                        if 'cloned' in meta:
-                            yield name
-                    else:
-                        yield name
-                except FileNotFoundError:
-                    pass
+    def all(cls):
+        pattern = f'{xdg_data_home}/virt-up/instance/*.json'
+        for path in glob.glob(pattern):
+            name = os.path.basename(path).replace('.json', '')
+            yield Instance(name)
 
     @classmethod
     def _domain_exists(cls, name):
@@ -586,7 +674,9 @@ class Instance:
               memory=512,
               size=None,
               vcpus=1,
-              graphics='none'):
+              graphics='none',
+              dns_domain=None,
+              **kwargs):
         """
         Build an instance with virt-builder and virt-install.
         """
@@ -621,30 +711,51 @@ class Instance:
         if os.path.exists(image):
             raise FileExistsError(f"Image file '{image}' already exists.")
 
-        # Generate the user creditials for login.
+        # Install virt-builder --run/--firstboot scripts.
+        for script in SCRIPTS:
+            path = os.path.join('/tmp/virt-up/scripts', script['name'])
+            if not os.path.exists(path):
+                if not os.path.exists('/tmp/virt-up/scripts'):
+                    os.makedirs('/tmp/virt-up/scripts')
+                with open(path, 'w') as f:
+                    f.write(script['contents'])
+                log.info('Wrote: %s', path)
+
+        # Generate the ssh keys. Generate the passwords if not provided.
         if not user:
             user = settings.username
         root_creds = Creds('root', password=root_password)
         user_creds = Creds(user, password=password)
 
+        # useradd requires a hashed password.
+        salt = crypt.mksalt(crypt.METHOD_SHA512)
+        password = crypt.crypt(user_creds.password, salt)
+
         # Setup virt-builder arguments.
+        if dns_domain is None:
+            dns_domain = settings.dns_domain
+        if dns_domain:
+            hostname = f'{name}.{dns_domain}'
+        else:
+            hostname = name
         extra_args = settings.extra_args('virt-builder')
         if size:
             extra_args.extend(['--size', size])
 
-        log.info(f"Building image file '{image}'.")
-        virt_builder(
-            settings.os_version,
-            '--output', image,
-            '--format', settings.image_format,
-            '--root-password', f'password:{root_creds.password}',
-            '--run-command', 'ssh-keygen -A',
-            '--run-command', f'id -u {user} || useradd -m -s /bin/bash -p "" {user}',
-            '--run-command', f'test -x /sbin/chpasswd && (echo "{user}:{user_creds.password}" | /sbin/chpasswd)',
-            '--ssh-inject', f'{user}:file:{user_creds.ssh_identity}.pub',
-            '--run-command', 'mkdir -p /etc/sudoers.d',
-            '--write',  f'/etc/sudoers.d/99-sna-devlab:{user} ALL=(ALL) NOPASSWD: ALL',
-            *extra_args)
+        with LockFile():
+            log.info(f"Building image file '{image}'.")
+            virt_builder(
+                settings.os_version,
+                '--output', image,
+                '--format', settings.image_format,
+                '--hostname', hostname,
+                '--root-password', f'password:{root_creds.password}',
+                '--run-command', 'ssh-keygen -A',
+                '--run-command', f"useradd -m -s /bin/bash -p '{password}' {user}",
+                '--ssh-inject', f'{user}:file:{user_creds.ssh_identity}.pub',
+                '--run-command', 'mkdir -p /etc/sudoers.d',
+                '--write',  f'/etc/sudoers.d/99-sna-devlab:{user} ALL=(ALL) NOPASSWD: ALL',
+                *extra_args)
 
         # Setup virt-install options. Reuse the last mac address for this
         # instance so it will (hopefully) be assigned the same address.
@@ -654,18 +765,19 @@ class Instance:
             optional_args.extend(['--mac', mac])
         extra_args = settings.extra_args('virt-install')
 
-        log.info(f"Importing instance '{name}'.")
-        virt_install(
-            '--import',
-            '--name', name,
-            '--disk', image,
-            '--memory', memory,
-            '--vcpus', vcpus,
-            '--graphics', graphics,
-            '--os-variant', settings.os_variant,
-            '--noautoconsole',
-            *optional_args,
-            *extra_args)
+        with LockFile():
+            log.info(f"Importing instance '{name}'.")
+            virt_install(
+                '--import',
+                '--name', name,
+                '--disk', image,
+                '--memory', memory,
+                '--vcpus', vcpus,
+                '--graphics', graphics,
+                '--os-variant', settings.os_variant,
+                '--noautoconsole',
+                *optional_args,
+                *extra_args)
 
         # Attach the new domain instance and update the meta data. Save the
         # assigned mac address for next time.
@@ -674,29 +786,37 @@ class Instance:
             'created': str(datetime.datetime.now()),
             'os_version': settings.os_version,
             'os_variant': settings.os_variant,
+            'hostname': hostname,
             'disk': image,
+            'image_format': settings.image_format,
             'memory': memory,
             'vcpus': vcpus,
             'graphics': graphics,
             'root': vars(root_creds),
             'user': vars(user_creds),
-            'address_source': settings.address_source,
+            'address-source': settings.address_source,
         }
         if size:
             meta['size'] = size
         instance = Instance(name, meta=meta)
         maddrs.update(name, instance.mac())
         instance.address() # Wait for address to be assigned.
+        Instance.update_inventory()
+        if settings.template_playbook:
+            instance.run_playbook(settings.template_playbook)
 
         return instance
 
     def clone(self,
             target,
             settings=None,
+            hostname=None,
             memory=None,
             size=None,
             vcpus=None,
             graphics=None,
+            dns_domain=None,
+            inventory=False,
             **kwargs):
         """
         Clone this instance to a new target instance.
@@ -731,28 +851,34 @@ class Instance:
         if os.path.exists(target_image):
             raise FileExistsError(f"Image file '{target_image}' already exists.")
         self.stop()  # Ensure we are stopped before cloning.
-        log.info(f"Cloning '{source_image}' to '{target_image}'.")
-        if settings.image_format == 'qcow2':
-            qemu_img.create('-f', 'qcow2', '-F', 'qcow2', '-b', source_image, target_image)
-        else:
-            extra_args = settings.extra_args('cp')
-            cp(*extra_args, source_image, target_image)
+
+        with LockFile():
+            log.info(f"Cloning '{source_image}' to '{target_image}'.")
+            if settings.image_format == 'qcow2':
+                qemu_img.create('-f', 'qcow2', '-F', 'qcow2', '-b', source_image, target_image)
+            else:
+                extra_args = settings.extra_args('cp')
+                cp(*extra_args, source_image, target_image)
 
         # Setup virt-sysprep args.
-        if settings.dns_domain:
-            hostname = f'{target}.{settings.dns_domain}'
-        else:
-            hostname = target
+        if not hostname:
+            if dns_domain is None:
+                dns_domain = settings.dns_domain
+            if dns_domain:
+                hostname = f'{target}.{dns_domain}'
+            else:
+                hostname = target
 
         extra_args = settings.extra_args('virt-sysprep')
 
-        log.info(f"Preparing target image '{target_image}'.")
-        virt_sysprep(
-            '--quiet',
-            '--add', target_image,
-            '--operations', 'defaults,-ssh-userdir',
-            '--hostname', hostname,
-            *extra_args)
+        with LockFile():
+            log.info(f"Preparing target image '{target_image}'.")
+            virt_sysprep(
+                '--quiet',
+                '--add', target_image,
+                '--operations', 'defaults,-ssh-userdir',
+                '--hostname', hostname,
+                *extra_args)
 
         # Setup virt-install options. Reuse the last mac address for this
         # instance so it will (hopefully) be assigned the same address.
@@ -770,37 +896,105 @@ class Instance:
 
         extra_args = settings.extra_args('virt-install')
 
-        log.info(f"Importing instance '{target}'.")
-        virt_install(
-            '--import',
-            '--name', target,
-            '--disk', target_image,
-            '--memory', memory,
-            '--vcpus', vcpus,
-            '--graphics', graphics,
-            '--os-variant', self.meta['os_variant'],
-            '--noautoconsole',
-            *optional_args,
-            *extra_args)
+        with LockFile():
+            log.info(f"Importing instance '{target}'.")
+            virt_install(
+                '--import',
+                '--name', target,
+                '--disk', target_image,
+                '--memory', memory,
+                '--vcpus', vcpus,
+                '--graphics', graphics,
+                '--os-variant', self.meta['os_variant'],
+                '--noautoconsole',
+                '--autostart',
+                *optional_args,
+                *extra_args)
 
         # Attach the new domain instance and update the meta data. Save the
         # assigned mac address for next time.
         meta = self.meta.copy()
         meta.pop('address', None)  # Remove the parent's address.
         meta['cloned'] = str(datetime.datetime.now())
+        meta['from'] = self.name
+        meta['hostname'] = hostname
         meta['disk'] = target_image
+        meta['format'] = settings.image_format
         meta['memory'] = memory
         meta['vcpus'] = vcpus
         meta['graphics'] = graphics
         instance = Instance(target, meta=meta)
         maddrs.update(target, instance.mac())
         instance.address() # Wait for an address to be assigned.
+        if inventory:
+            Instance.update_inventory()
+            if settings.instance_playbook:
+                instance.run_playbook(settings.instance_playbook)
+
         return instance
 
-    def login(self, command=None):
+    @classmethod
+    def update_inventory(cls):
         """
-        ssh login to the instance.
+        Create an ansible inventory file for the cloned instances.
         """
+        filename = f'{xdg_data_home}/virt-up/inventory.yaml'
+        ssh_options = [
+            ('ControlMaster', 'auto'),
+            ('ControlPersist', '60s'),
+            ('ForwardX11', 'no'),
+            ('IdentitiesOnly', 'yes'),
+            ('LogLevel', 'ERROR'),
+            ('StrictHostKeyChecking', 'no'),
+            ('UserKnownHostsFile', '/dev/null'),
+        ]
+        ssh_common_args = ' '.join(['-o %s=%s' % x for x in ssh_options])
+        clones = {}
+        templates = {}
+        for instance in Instance.all():
+            name = instance.name
+            address = instance.meta.get('address', None)
+            if not address:
+                log.warning(f"Skipping inventory entry for instance '{name}'; address is not available.")
+                continue
+            host = {
+                'ansible_user': instance.meta['user']['username'],
+                'ansible_host': address,
+                'ansible_port': '22',
+                'ansible_private_key_file': instance.meta['user']['ssh_identity'],
+                'ansible_connection': 'ssh',
+                'ansible_ssh_common_args': ssh_common_args,
+            }
+            if instance.is_clone():
+                clones[name] = host
+            else:
+                templates[name] = host
+        with open(filename, 'w') as fp:
+            fp.writelines([
+                '---\n',
+                'all:\n',
+                '  children:\n',
+                '    virt_up_managed:\n',
+                '      hosts:\n'])
+            for name in clones.keys():
+                fp.write(f'        {name}:\n')
+                for key, value in clones[name].items():
+                    fp.write(f'          {key}: "{value}"\n')
+            fp.writelines([
+                '    virt_up_templates:\n',
+                '      hosts:\n'])
+            for name in templates.keys():
+                fp.write(f'        {name}:\n')
+                for key, value in templates[name].items():
+                    fp.write(f'          {key}: "{value}"\n')
+
+    def login(self, mode='ssh', command=None):
+        """
+        ssh or stfp login to the instance.
+        """
+        modes = {'ssh': ssh.__name__, 'sftp': sftp.__name__}
+        if mode not in modes:
+            raise ValueError(f"Unsupported mode '{mode}'.")
         self.start()
         self.meta.pop('address', None) # Flush our cached address.
         address = self.address()
@@ -815,19 +1009,59 @@ class Instance:
             '-o', 'LogLevel=ERROR',
             f'{user}@{address}',
         ]
-
-        self.wait_for_port(22)  # Wait until ssh port is ready.
-        if not command:
-            args.insert(0, ssh.__name__) # Required for execv.
-            os.execv(ssh.__name__, args) # Drop into interactive shell, never to return.
-            assert(False) # unreachable
-        else:
+        if command:
             args.append(command)
-            output = []
-            try:
-                for line in ssh(*args, _err=logerr, _iter=True):
-                    logout(line)
-                    output.append(line.rstrip())
-            except sh.ErrorReturnCode as e:
-                log.error(e)
-            return ''.join(output)
+
+        self.wait_for_port(22)      # Wait until ssh port is ready.
+        args.insert(0, modes[mode]) # Required for execv.
+        os.execv(modes[mode], args) # Drop into interactive shell, never to return.
+        raise AssertionError('exec failed')
+
+    def run_command(self, *args, sudo=False):
+        """
+        Run a command via ssh and return the exit code, stdout,
+        and stderr as a tuple.
+        """
+        self.wait_for_port(22)
+        address = self.address()
+        user = self.meta['user']['username']
+        ssh_identity = self.meta['user']['ssh_identity']
+        if sudo:
+            args = ['sudo', '-n'] + list(args)
+        command = shlex.join(args)
+        ssh_args = [
+            '-i', ssh_identity,
+            '-o', 'PasswordAuthentication=no',
+            '-o', 'CheckHostIP=no',
+            '-o', 'UserKnownHostsFile=/dev/null',
+            '-o', 'StrictHostKeyChecking=no',
+            '-o', 'LogLevel=ERROR',
+            f'{user}@{address}',
+            command,
+        ]
+        code = 0
+        out = io.StringIO()
+        err = io.StringIO()
+        try:
+            ssh(ssh_args, _out=out, _err=err)
+        except sh.ErrorReturnCode as e:
+            code = e.exit_code
+        return code, out.getvalue(), err.getvalue()
+
+    def run_playbook(self, playbook):
+        """
+        Run an ansible playbook on this instance.
+        """
+        inventory = f'{xdg_data_home}/virt-up/inventory.yaml'
+        address = self.address()
+        if not address:
+            log.warning(f"Skipping playbook; address for '{self.name}' is not available.")
+            return
+        if not os.path.exists(playbook):
+            log.warning(f"Skipping playbook; '{playbook}' file not found.")
+            return
+        if not ansible:
+            log.warning("Skipping playbook; 'ansible-playbook' command not found.")
+            return
+        log.info(f"Running playbook '{playbook}'.")
+        ansible('-i', inventory, '--limit', self.name, playbook)
